@@ -51,46 +51,54 @@ the form
 
   language[_TERRITORY][.charset][@@modifier]
 
-<<"language">> is a two character string per ISO 639.  <<"TERRITORY">> is a
-country code per ISO 3166.  For <<"charset">> and <<"modifier">> see below.
+<<"language">> is a two character string per ISO 639, or, if not available
+for a given language, a three character string per ISO 639-3.
+<<"TERRITORY">> is a country code per ISO 3166.  For <<"charset">> and
+<<"modifier">> see below.
 
-Additionally to the POSIX specifier, seven extensions are supported for
-backward compatibility with older implementations using newlib:
-<<"C-UTF-8">>, <<"C-JIS">>, <<"C-eucJP">>, <<"C-SJIS">>, <<C-KOI8-R>>,
-<<C-KOI8-U>>, <<"C-ISO-8859-x">> with 1 <= x <= 15, or <<"C-CPxxx">> with
-xxx in [437, 720, 737, 775, 850, 852, 855, 857, 858, 862, 866, 874, 1125,
-1250, 1251, 1252, 1253, 1254, 1255, 1256, 1257, 1258].
-
-Instead of <<"C-">>, you can specify also <<"C.">>.  Both variations allow
+Additionally to the POSIX specifier, the following extension is supported
+for backward compatibility with older implementations using newlib:
+<<"C-charset">>.
+Instead of <<"C-">>, you can also specify <<"C.">>.  Both variations allow
 to specify language neutral locales while using other charsets than ASCII,
 for instance <<"C.UTF-8">>, which keeps all settings as in the C locale,
 but uses the UTF-8 charset.
 
-Even when using POSIX locale strings, the only charsets allowed are
-<<"UTF-8">>, <<"JIS">>, <<"EUCJP">>, <<"SJIS">>, <<KOI8-R>>, <<KOI8-U>>,
-<<"ISO-8859-x">> with 1 <= x <= 15, or <<"CPxxx">> with xxx in
-[437, 720, 737, 775, 850, 852, 855, 857, 858, 862, 866, 874, 1125, 1250,
-1251, 1252, 1253, 1254, 1255, 1256, 1257, 1258].
+The following charsets are recognized:
+<<"UTF-8">>, <<"JIS">>, <<"EUCJP">>, <<"SJIS">>, <<"KOI8-R">>, <<"KOI8-U">>,
+<<"GEORGIAN-PS">>, <<"PT154">>, <<"TIS-620">>, <<"ISO-8859-x">> with
+1 <= x <= 16, or <<"CPxxx">> with xxx in [437, 720, 737, 775, 850, 852, 855,
+857, 858, 862, 866, 874, 932, 1125, 1250, 1251, 1252, 1253, 1254, 1255, 1256,
+1257, 1258].
+
 Charsets are case insensitive.  For instance, <<"EUCJP">> and <<"eucJP">>
-are equivalent.  <<"UTF-8">> can also be written without dash, as in
-<<"UTF8">> or <<"utf8">>.
+are equivalent.  Charset names with dashes can also be written without
+dashes, as in <<"UTF8">>, <<"iso88591">> or <<"koi8r">>.  <<"EUCJP">> and
+<<"EUCKR">> are also recognized with dash, <<"EUC-JP">> and <<"EUC-KR">>.
+
+Full support for all of the above charsets requires that newlib has been
+build with multibyte support and support for all ISO and Windows Codepage.
+Otherwise all singlebyte charsets are simply mapped to ASCII.  Right now,
+only newlib for Cygwin is built with full charset support by default.
+Under Cygwin, this implementation additionally supports the charsets
+<<"GBK">>, <<"GB2312">>, <<"eucCN">>, <<"eucKR">>, and <<"Big5">>.  Cygwin
+does not support <<"JIS">>.
+
+Cygwin additionally supports locales from the file
+/usr/share/locale/locale.alias.
 
 (<<"">> is also accepted; if given, the settings are read from the
-corresponding LC_* environment variables and $LANG according to POSIX rules.
+corresponding LC_* environment variables and $LANG according to POSIX rules.)
 
-Under Cygwin, this implementation additionally supports the charsets
-<<"GBK">>, <<"eucKR">>, and <<"Big5">>.
-
-This implementation also supports a single modifier, <<"cjknarrow">>.
-Any other modifier is ignored.  <<"cjknarrow">>, in conjunction with one
-of the language specifiers <<"ja">>, <<"ko">>, and <<"zh">> specifies
-how the functions <<wcwidth>> and <<wcswidth>> handle characters from
-the "CJK Ambiguous Width" character class described in
-http://www.unicode.org/unicode/reports/tr11/.  Usually these characters
-have a width of 1, unless you specify one of the aforementioned
-languages, in which case these characters have a width of 2.  By
-specifying the <<"cjknarrow">> modifier, these characters will have a
-width of one in the languages <<"ja">>, <<"ko">>, and <<"zh">> as well.
+This implementation also supports the modifier <<"cjknarrow">>, which
+affects how the functions <<wcwidth>> and <<wcswidth>> handle characters
+from the "CJK Ambiguous Width" category of characters described at
+http://www.unicode.org/reports/tr11/#Ambiguous. These characters have a width
+of 1 for singlebyte charsets and a width of 2 for multibyte charsets
+other than UTF-8. For UTF-8, their width depends on the language specifier:
+it is 2 for <<"zh">> (Chinese), <<"ja">> (Japanese), and <<"ko">> (Korean),
+and 1 for everything else. Specifying <<"cjknarrow">> forces a width of 1,
+independent of charset and language.
 
 If you use <<NULL>> as the <[locale]> argument, <<setlocale>> returns a
 pointer to the string representing the current locale.  The acceptable
@@ -164,32 +172,24 @@ No supporting OS subroutines are required.
 
 #include <newlib.h>
 #include <errno.h>
-#include <locale.h>
 #include <string.h>
 #include <limits.h>
 #include <reent.h>
 #include <stdlib.h>
 #include <wchar.h>
+#include "setlocale.h"
+#include "../ctype/ctype_.h"
 #include "../stdlib/local.h"
 
-#define _LC_LAST      7
-#define ENCODING_LEN 31
+#ifdef __CYGWIN__ /* Has to be kept available as exported symbol for
+		     backward compatibility.  Set it in setlocale, but
+		     otherwise ignore it.  Applications compiled after
+		     2010 don't use it anymore. */
+int __EXPORT __mb_cur_max = 6;
+#endif
 
-int __EXPORT __mb_cur_max = 1;
-
-int __nlocale_changed = 0;
-int __mlocale_changed = 0;
 char *_PathLocale = NULL;
 
-static _CONST struct lconv lconv = 
-{
-  ".", "", "", "", "", "", "", "", "", "",
-  CHAR_MAX, CHAR_MAX, CHAR_MAX, CHAR_MAX,
-  CHAR_MAX, CHAR_MAX, CHAR_MAX, CHAR_MAX,
-  CHAR_MAX, CHAR_MAX, CHAR_MAX, CHAR_MAX,
-  CHAR_MAX, CHAR_MAX
-};
-
 #ifdef _MB_CAPABLE
 /*
  * Category names for getenv()
@@ -203,6 +203,7 @@ static char *categories[_LC_LAST] = {
   "LC_TIME",
   "LC_MESSAGES",
 };
+#endif /* _MB_CAPABLE */
 
 /*
  * Default locale per POSIX.  Can be overridden on a per-target base.
@@ -210,46 +211,98 @@ static char *categories[_LC_LAST] = {
 #ifndef DEFAULT_LOCALE
 #define DEFAULT_LOCALE	"C"
 #endif
+
+#ifdef _MB_CAPABLE
 /*
  * This variable can be changed by any outside mechanism.  This allows,
  * for instance, to load the default locale from a file.
  */
 char __default_locale[ENCODING_LEN + 1] = DEFAULT_LOCALE;
 
-/*
- * Current locales for each category
- */
-static char current_categories[_LC_LAST][ENCODING_LEN + 1] = {
-    "C",
-    "C",
-    "C",
-    "C",
-    "C",
-    "C",
-    "C",
+const struct __locale_t __C_locale =
+{
+  { "C", "C", "C", "C", "C", "C", "C", },
+  __ascii_wctomb,
+  __ascii_mbtowc,
+  0,
+  DEFAULT_CTYPE_PTR,
+  {
+    ".", "", "", "", "", "", "", "", "", "",
+    CHAR_MAX, CHAR_MAX, CHAR_MAX, CHAR_MAX,
+    CHAR_MAX, CHAR_MAX, CHAR_MAX, CHAR_MAX,
+    CHAR_MAX, CHAR_MAX, CHAR_MAX, CHAR_MAX,
+    CHAR_MAX, CHAR_MAX
+  },
+#ifndef __HAVE_LOCALE_INFO__
+  "\1",
+  "ASCII",
+  "ASCII",
+#else /* __HAVE_LOCALE_INFO__ */
+  {
+    { NULL, NULL },			/* LC_ALL */
+#ifdef __CYGWIN__
+    { &_C_collate_locale, NULL },	/* LC_COLLATE */
+#else
+    { NULL, NULL },			/* LC_COLLATE */
+#endif
+    { &_C_ctype_locale, NULL },		/* LC_CTYPE */
+    { &_C_monetary_locale, NULL },	/* LC_MONETARY */
+    { &_C_numeric_locale, NULL },	/* LC_NUMERIC */
+    { &_C_time_locale, NULL },		/* LC_TIME */
+    { &_C_messages_locale, NULL },	/* LC_MESSAGES */
+  },
+#endif /* __HAVE_LOCALE_INFO__ */
+};
+#endif /* _MB_CAPABLE */
+
+struct __locale_t __global_locale =
+{
+  { "C", "C", DEFAULT_LOCALE, "C", "C", "C", "C", },
+#ifdef __CYGWIN__
+  __utf8_wctomb,
+  __utf8_mbtowc,
+#else
+  __ascii_wctomb,
+  __ascii_mbtowc,
+#endif
+  0,
+  DEFAULT_CTYPE_PTR,
+  {
+    ".", "", "", "", "", "", "", "", "", "",
+    CHAR_MAX, CHAR_MAX, CHAR_MAX, CHAR_MAX,
+    CHAR_MAX, CHAR_MAX, CHAR_MAX, CHAR_MAX,
+    CHAR_MAX, CHAR_MAX, CHAR_MAX, CHAR_MAX,
+    CHAR_MAX, CHAR_MAX
+  },
+#ifndef __HAVE_LOCALE_INFO__
+  "\1",
+  "ASCII",
+  "ASCII",
+#else /* __HAVE_LOCALE_INFO__ */
+  {
+    { NULL, NULL },			/* LC_ALL */
+#ifdef __CYGWIN__
+    { &_C_collate_locale, NULL },	/* LC_COLLATE */
+#else
+    { NULL, NULL },			/* LC_COLLATE */
+#endif
+    { &_C_ctype_locale, NULL },		/* LC_CTYPE */
+    { &_C_monetary_locale, NULL },	/* LC_MONETARY */
+    { &_C_numeric_locale, NULL },	/* LC_NUMERIC */
+    { &_C_time_locale, NULL },		/* LC_TIME */
+    { &_C_messages_locale, NULL },	/* LC_MESSAGES */
+  },
+#endif /* __HAVE_LOCALE_INFO__ */
 };
 
-/*
- * The locales we are going to try and load
- */
-static char new_categories[_LC_LAST][ENCODING_LEN + 1];
-static char saved_categories[_LC_LAST][ENCODING_LEN + 1];
+#ifdef _MB_CAPABLE
+/* Renamed from current_locale_string to make clear this is only the
+   *global* string for setlocale (LC_ALL, NULL).  There's no equivalent
+   functionality for uselocale. */
+static char global_locale_string[_LC_LAST * (ENCODING_LEN + 1/*"/"*/ + 1)];
+static char *currentlocale (void);
 
-static char current_locale_string[_LC_LAST * (ENCODING_LEN + 1/*"/"*/ + 1)];
-static char *currentlocale(void);
-static char *loadlocale(struct _reent *, int);
-static const char *__get_locale_env(struct _reent *, int);
-
-#endif
-
-#ifdef __CYGWIN__
-static char lc_ctype_charset[ENCODING_LEN + 1] = "UTF-8";
-static char lc_message_charset[ENCODING_LEN + 1] = "UTF-8";
-#else
-static char lc_ctype_charset[ENCODING_LEN + 1] = "ASCII";
-static char lc_message_charset[ENCODING_LEN + 1] = "ASCII";
-#endif
-static int lc_ctype_cjk_lang = 0;
+#endif /* _MB_CAPABLE */
 
 char *
 _DEFUN(_setlocale_r, (p, category, locale),
@@ -265,7 +318,9 @@ _DEFUN(_setlocale_r, (p, category, locale),
         return NULL;
     }
   return "C";
-#else
+#else /* _MB_CAPABLE */
+  static char new_categories[_LC_LAST][ENCODING_LEN + 1];
+  static char saved_categories[_LC_LAST][ENCODING_LEN + 1];
   int i, j, len, saverr;
   const char *env, *r;
 
@@ -276,13 +331,14 @@ _DEFUN(_setlocale_r, (p, category, locale),
     }
 
   if (locale == NULL)
-    return category != LC_ALL ? current_categories[category] : currentlocale();
+    return category != LC_ALL ? __get_global_locale ()->categories[category]
+			      : currentlocale();
 
   /*
    * Default to the current locale for everything.
    */
   for (i = 1; i < _LC_LAST; ++i)
-    strcpy (new_categories[i], current_categories[i]);
+    strcpy (new_categories[i], __get_global_locale ()->categories[i]);
 
   /*
    * Now go fill up new_categories from the locale argument
@@ -370,21 +426,23 @@ _DEFUN(_setlocale_r, (p, category, locale),
     }
 
   if (category != LC_ALL)
-    return loadlocale (p, category);
+    return __loadlocale (__get_global_locale (), category,
+			 new_categories[category]);
 
   for (i = 1; i < _LC_LAST; ++i)
     {
-      strcpy (saved_categories[i], current_categories[i]);
-      if (loadlocale (p, i) == NULL)
+      strcpy (saved_categories[i], __get_global_locale ()->categories[i]);
+      if (__loadlocale (__get_global_locale (), i, new_categories[i]) == NULL)
 	{
 	  saverr = p->_errno;
 	  for (j = 1; j < i; j++)
 	    {
 	      strcpy (new_categories[j], saved_categories[j]);
-	      if (loadlocale (p, j) == NULL)
+	      if (__loadlocale (__get_global_locale (), j, new_categories[j])
+		  == NULL)
 		{
 		  strcpy (new_categories[j], "C");
-		  loadlocale (p, j);
+		  __loadlocale (__get_global_locale (), j, new_categories[j]);
 		}
 	    }
 	  p->_errno = saverr;
@@ -392,39 +450,36 @@ _DEFUN(_setlocale_r, (p, category, locale),
 	}
     }
   return currentlocale ();
-#endif
+#endif /* _MB_CAPABLE */
 }
 
 #ifdef _MB_CAPABLE
 static char *
-currentlocale()
+currentlocale ()
 {
-        int i;
+  int i;
 
-        (void)strcpy(current_locale_string, current_categories[1]);
+  strcpy (global_locale_string, __get_global_locale ()->categories[1]);
 
-        for (i = 2; i < _LC_LAST; ++i)
-                if (strcmp(current_categories[1], current_categories[i])) {
-                        for (i = 2; i < _LC_LAST; ++i) {
-                                (void)strcat(current_locale_string, "/");
-                                (void)strcat(current_locale_string,
-                                             current_categories[i]);
-                        }
-                        break;
-                }
-        return (current_locale_string);
+  for (i = 2; i < _LC_LAST; ++i)
+    if (strcmp (__get_global_locale ()->categories[1],
+		__get_global_locale ()->categories[i]))
+      {
+	for (i = 2; i < _LC_LAST; ++i)
+	  {
+	    (void)strcat(global_locale_string, "/");
+	    (void)strcat(global_locale_string,
+			 __get_global_locale ()->categories[i]);
+	  }
+	break;
+      }
+  return global_locale_string;
 }
-#endif
 
-#ifdef _MB_CAPABLE
-#ifdef __CYGWIN__
-extern void *__set_charset_from_codepage (unsigned int, char *charset);
-#endif /* __CYGWIN__ */
+extern void __set_ctype (struct __locale_t *, const char *charset);
 
-extern void __set_ctype (const char *charset);
-
-static char *
-loadlocale(struct _reent *p, int category)
+char *
+__loadlocale (struct __locale_t *loc, int category, const char *new_locale)
 {
   /* At this point a full-featured system would just load the locale
      specific data from the locale files.
@@ -432,29 +487,54 @@ loadlocale(struct _reent *p, int category)
      The string must be in one of the allowed locale strings, either
      one in POSIX-style, or one in the old newlib style to maintain
      backward compatibility.  If the local string is correct, the charset
-     is extracted and stored in lc_ctype_charset or lc_message_charset
+     is extracted and stored in ctype_codeset or message_charset
      dependent on the cateogry. */
-  char *locale = new_categories[category];
+  char *locale = NULL;
   char charset[ENCODING_LEN + 1];
-  unsigned long val;
-  char *end;
+  long val = 0;
+  char *end, *c = NULL;
   int mbc_max;
-  int (*l_wctomb) (struct _reent *, char *, wchar_t, const char *, mbstate_t *);
-  int (*l_mbtowc) (struct _reent *, wchar_t *, const char *, size_t,
-		   const char *, mbstate_t *);
-#ifdef _MB_CAPABLE
+  wctomb_p l_wctomb;
+  mbtowc_p l_mbtowc;
   int cjknarrow = 0;
+
+  /* Avoid doing everything twice if nothing has changed. */
+  if (!strcmp (new_locale, loc->categories[category]))
+    return loc->categories[category];
+
+#ifdef __CYGWIN__
+  /* This additional code handles the case that the incoming locale string
+     is not valid.  If so, it calls the function __set_locale_from_locale_alias,
+     which is only available on Cygwin right now.  The function reads the
+     file /usr/share/locale/locale.alias.  The file contains locale aliases
+     and their replacement locale.  For instance, the alias "french" is
+     translated to "fr_FR.ISO-8859-1", the alias "thai" is translated to
+     "th_TH.TIS-620".  If successful, the function returns with a pointer
+     to the second argument, which is a buffer in which the replacement locale
+     gets stored.  Otherwise the function returns NULL. */
+  char tmp_locale[ENCODING_LEN + 1];
+  int ret = 0;
+
+restart:
+  if (!locale)
+    locale = (char *) new_locale;
+  else if (locale != tmp_locale)
+    {
+      locale = __set_locale_from_locale_alias (locale, tmp_locale);
+      if (!locale)
+	return NULL;
+    }
+# define FAIL	goto restart
+#else
+  locale = new_locale;
+# define FAIL	return NULL
 #endif
-  
+
   /* "POSIX" is translated to "C", as on Linux. */
   if (!strcmp (locale, "POSIX"))
     strcpy (locale, "C");
   if (!strcmp (locale, "C"))				/* Default "C" locale */
-#ifdef __CYGWIN__
-    strcpy (charset, "UTF-8");
-#else
     strcpy (charset, "ASCII");
-#endif
   else if (locale[0] == 'C'
 	   && (locale[1] == '-'		/* Old newlib style */
 	       || locale[1] == '.'))	/* Extension for the C locale to allow
@@ -462,24 +542,36 @@ loadlocale(struct _reent *p, int category)
 					   sticking to the C locale in terms
 					   of sort order, etc.  Proposed in
 					   the Debian project. */
-    strcpy (charset, locale + 2);
+    {
+      char *chp;
+
+      c = locale + 2;
+      strcpy (charset, c);
+      if ((chp = strchr (charset, '@')))
+        /* Strip off modifier */
+        *chp = '\0';
+      c += strlen (charset);
+    }
   else							/* POSIX style */
     {
-      char *c = locale;
+      c = locale;
 
       /* Don't use ctype macros here, they might be localized. */
       /* Language */
       if (c[0] < 'a' || c[0] > 'z'
 	  || c[1] < 'a' || c[1] > 'z')
-	return NULL;
+	FAIL;
       c += 2;
+      /* Allow three character Language per ISO 639-3 */
+      if (c[0] >= 'a' && c[0] <= 'z')
+      	++c;
       if (c[0] == '_')
         {
 	  /* Territory */
 	  ++c;
 	  if (c[0] < 'A' || c[0] > 'Z'
 	      || c[1] < 'A' || c[1] > 'Z')
-	    return NULL;
+	    FAIL;
 	  c += 2;
 	}
       if (c[0] == '.')
@@ -497,24 +589,28 @@ loadlocale(struct _reent *p, int category)
       else if (c[0] == '\0' || c[0] == '@')
 	/* End of string or just a modifier */
 #ifdef __CYGWIN__
-	__set_charset_from_codepage (0, charset);
+	/* The Cygwin-only function __set_charset_from_locale checks
+	   for the default charset which is connected to the given locale.
+	   The function uses Windows functions in turn so it can't be easily
+	   adapted to other targets.  However, if any other target provides
+	   equivalent functionality, preferrably using the same function name
+	   it would be sufficient to change the guarding #ifdef. */
+	__set_charset_from_locale (locale, charset);
 #else
 	strcpy (charset, "ISO-8859-1");
 #endif
       else
 	/* Invalid string */
-      	return NULL;
-#ifdef _MB_CAPABLE
-      if (c[0] == '@')
-	{
-	  /* Modifier */
-	  /* Only one modifier is recognized right now.  "cjknarrow" is used
-	     to modify the behaviour of wcwidth() for East Asian languages.
-	     For details see the comment at the end of this function. */
-	  if (!strcmp (c + 1, "cjknarrow"))
-	    cjknarrow = 1;
-	}
-#endif
+      	FAIL;
+    }
+  if (c && c[0] == '@')
+    {
+      /* Modifier */
+      /* Only one modifier is recognized right now.  "cjknarrow" is used
+         to modify the behaviour of wcwidth() for East Asian languages.
+         For details see the comment at the end of this function. */
+      if (!strcmp (c + 1, "cjknarrow"))
+	cjknarrow = 1;
     }
   /* We only support this subset of charsets. */
   switch (charset[0])
@@ -522,90 +618,108 @@ loadlocale(struct _reent *p, int category)
     case 'U':
     case 'u':
       if (strcasecmp (charset, "UTF-8") && strcasecmp (charset, "UTF8"))
-	return NULL;
+	FAIL;
       strcpy (charset, "UTF-8");
       mbc_max = 6;
-#ifdef _MB_CAPABLE
       l_wctomb = __utf8_wctomb;
       l_mbtowc = __utf8_mbtowc;
-#endif
     break;
+#ifndef __CYGWIN__
+    /* Cygwin does not support JIS at all. */
     case 'J':
     case 'j':
       if (strcasecmp (charset, "JIS"))
-	return NULL;
+	FAIL;
       strcpy (charset, "JIS");
       mbc_max = 8;
-#ifdef _MB_CAPABLE
       l_wctomb = __jis_wctomb;
       l_mbtowc = __jis_mbtowc;
-#endif
     break;
+#endif /* !__CYGWIN__ */
     case 'E':
     case 'e':
-      if (!strcasecmp (charset, "EUCJP"))
+      if (strncasecmp (charset, "EUC", 3))
+	FAIL;
+      c = charset + 3;
+      if (*c == '-')
+	++c;
+      if (!strcasecmp (c, "JP"))
 	{
 	  strcpy (charset, "EUCJP");
 	  mbc_max = 3;
-#ifdef _MB_CAPABLE
 	  l_wctomb = __eucjp_wctomb;
 	  l_mbtowc = __eucjp_mbtowc;
-#endif
 	}
 #ifdef __CYGWIN__
-      else if (!strcasecmp (charset, "EUCKR"))
+      /* Newlib does neither provide EUC-KR nor EUC-CN, and Cygwin's
+      	 implementation requires Windows support. */
+      else if (!strcasecmp (c, "KR"))
 	{
 	  strcpy (charset, "EUCKR");
 	  mbc_max = 2;
-#ifdef _MB_CAPABLE
 	  l_wctomb = __kr_wctomb;
 	  l_mbtowc = __kr_mbtowc;
-#endif
 	}
-#endif
+      else if (!strcasecmp (c, "CN"))
+	{
+	  strcpy (charset, "EUCCN");
+	  mbc_max = 2;
+	  l_wctomb = __gbk_wctomb;
+	  l_mbtowc = __gbk_mbtowc;
+	}
+#endif /* __CYGWIN__ */
       else
-	return NULL;
+	FAIL;
     break;
     case 'S':
     case 's':
       if (strcasecmp (charset, "SJIS"))
-	return NULL;
+	FAIL;
       strcpy (charset, "SJIS");
       mbc_max = 2;
-#ifdef _MB_CAPABLE
       l_wctomb = __sjis_wctomb;
       l_mbtowc = __sjis_mbtowc;
-#endif
     break;
     case 'I':
     case 'i':
       /* Must be exactly one of ISO-8859-1, [...] ISO-8859-16, except for
-         ISO-8859-12. */
-      if (strncasecmp (charset, "ISO-8859-", 9))
-	return NULL;
-      strncpy (charset, "ISO", 3);
-      val = _strtol_r (p, charset + 9, &end, 10);
+         ISO-8859-12.  This code also recognizes the aliases without dashes. */
+      if (strncasecmp (charset, "ISO", 3))
+	FAIL;
+      c = charset + 3;
+      if (*c == '-')
+	++c;
+      if (strncasecmp (c, "8859", 4))
+	FAIL;
+      c += 4;
+      if (*c == '-')
+	++c;
+      val = strtol (c, &end, 10);
       if (val < 1 || val > 16 || val == 12 || *end)
-	return NULL;
+	FAIL;
+      strcpy (charset, "ISO-8859-");
+      c = charset + 9;
+      if (val > 10)
+      	*c++ = '1';
+      *c++ = val % 10 + '0';
+      *c = '\0';
       mbc_max = 1;
-#ifdef _MB_CAPABLE
 #ifdef _MB_EXTENDED_CHARSETS_ISO
-      l_wctomb = __iso_wctomb;
-      l_mbtowc = __iso_mbtowc;
+      l_wctomb = __iso_wctomb (val);
+      l_mbtowc = __iso_mbtowc (val);
 #else /* !_MB_EXTENDED_CHARSETS_ISO */
       l_wctomb = __ascii_wctomb;
       l_mbtowc = __ascii_mbtowc;
 #endif /* _MB_EXTENDED_CHARSETS_ISO */
-#endif
     break;
     case 'C':
     case 'c':
       if (charset[1] != 'P' && charset[1] != 'p')
-	return NULL;
+	FAIL;
       strncpy (charset, "CP", 2);
-      val = _strtol_r (p, charset + 2, &end, 10);
+      val = strtol (charset + 2, &end, 10);
       if (*end)
-	return NULL;
+	FAIL;
       switch (val)
 	{
 	case 437:
@@ -631,104 +745,216 @@ loadlocale(struct _reent *p, int category)
 	case 1257:
 	case 1258:
 	  mbc_max = 1;
-#ifdef _MB_CAPABLE
 #ifdef _MB_EXTENDED_CHARSETS_WINDOWS
-	  l_wctomb = __cp_wctomb;
-	  l_mbtowc = __cp_mbtowc;
+	  l_wctomb = __cp_wctomb (val);
+	  l_mbtowc = __cp_mbtowc (val);
 #else /* !_MB_EXTENDED_CHARSETS_WINDOWS */
 	  l_wctomb = __ascii_wctomb;
 	  l_mbtowc = __ascii_mbtowc;
 #endif /* _MB_EXTENDED_CHARSETS_WINDOWS */
-#endif
+	  break;
+	case 932:
+	  mbc_max = 2;
+	  l_wctomb = __sjis_wctomb;
+	  l_mbtowc = __sjis_mbtowc;
 	  break;
 	default:
-	  return NULL;
+	  FAIL;
 	}
     break;
     case 'K':
     case 'k':
-      if (!strcasecmp (charset, "KOI8-R"))
-	strcpy (charset, "CP20866");
-      else if (!strcasecmp (charset, "KOI8-U"))
-	strcpy (charset, "CP21866");
+      /* KOI8-R, KOI8-U and the aliases without dash */
+      if (strncasecmp (charset, "KOI8", 4))
+	FAIL;
+      c = charset + 4;
+      if (*c == '-')
+	++c;
+      if (*c == 'R' || *c == 'r')
+	{
+	  val = 20866;
+	  strcpy (charset, "CP20866");
+	}
+      else if (*c == 'U' || *c == 'u')
+	{
+	  val = 21866;
+	  strcpy (charset, "CP21866");
+	}
       else
-	return NULL;
+	FAIL;
       mbc_max = 1;
-#ifdef _MB_CAPABLE
 #ifdef _MB_EXTENDED_CHARSETS_WINDOWS
-      l_wctomb = __cp_wctomb;
-      l_mbtowc = __cp_mbtowc;
+      l_wctomb = __cp_wctomb (val);
+      l_mbtowc = __cp_mbtowc (val);
 #else /* !_MB_EXTENDED_CHARSETS_WINDOWS */
       l_wctomb = __ascii_wctomb;
       l_mbtowc = __ascii_mbtowc;
 #endif /* _MB_EXTENDED_CHARSETS_WINDOWS */
-#endif
       break;
     case 'A':
     case 'a':
       if (strcasecmp (charset, "ASCII"))
-	return NULL;
+	FAIL;
       strcpy (charset, "ASCII");
       mbc_max = 1;
-#ifdef _MB_CAPABLE
       l_wctomb = __ascii_wctomb;
       l_mbtowc = __ascii_mbtowc;
-#endif
       break;
-#ifdef __CYGWIN__
     case 'G':
     case 'g':
-      if (strcasecmp (charset, "GBK"))
-      	return NULL;
-      strcpy (charset, "GBK");
-      mbc_max = 2;
-#ifdef _MB_CAPABLE
-      l_wctomb = __gbk_wctomb;
-      l_mbtowc = __gbk_mbtowc;
-#endif
+#ifdef __CYGWIN__
+      /* Newlib does not provide GBK/GB2312 and Cygwin's implementation
+	 requires Windows support. */
+      if (!strcasecmp (charset, "GBK")
+	  || !strcasecmp (charset, "GB2312"))
+      	{
+	  strcpy (charset, charset[2] == '2' ? "GB2312" : "GBK");
+	  mbc_max = 2;
+	  l_wctomb = __gbk_wctomb;
+	  l_mbtowc = __gbk_mbtowc;
+	}
+      else
+#endif /* __CYGWIN__ */
+      /* GEORGIAN-PS and the alias without dash */
+      if (!strncasecmp (charset, "GEORGIAN", 8))
+	{
+	  c = charset + 8;
+	  if (*c == '-')
+	    ++c;
+	  if (strcasecmp (c, "PS"))
+	    FAIL;
+	  val = 101;
+	  strcpy (charset, "CP101");
+	  mbc_max = 1;
+#ifdef _MB_EXTENDED_CHARSETS_WINDOWS
+	  l_wctomb = __cp_wctomb (val);
+	  l_mbtowc = __cp_mbtowc (val);
+#else /* !_MB_EXTENDED_CHARSETS_WINDOWS */
+	  l_wctomb = __ascii_wctomb;
+	  l_mbtowc = __ascii_mbtowc;
+#endif /* _MB_EXTENDED_CHARSETS_WINDOWS */
+	}
+      else
+	FAIL;
       break;
+    case 'P':
+    case 'p':
+      /* PT154 */
+      if (strcasecmp (charset, "PT154"))
+	FAIL;
+      val = 102;
+      strcpy (charset, "CP102");
+      mbc_max = 1;
+#ifdef _MB_EXTENDED_CHARSETS_WINDOWS
+      l_wctomb = __cp_wctomb (val);
+      l_mbtowc = __cp_mbtowc (val);
+#else /* !_MB_EXTENDED_CHARSETS_WINDOWS */
+      l_wctomb = __ascii_wctomb;
+      l_mbtowc = __ascii_mbtowc;
+#endif /* _MB_EXTENDED_CHARSETS_WINDOWS */
+      break;
+    case 'T':
+    case 't':
+      if (strncasecmp (charset, "TIS", 3))
+      	FAIL;
+      c = charset + 3;
+      if (*c == '-')
+	++c;
+      if (strcasecmp (c, "620"))
+      	FAIL;
+      val = 874;
+      strcpy (charset, "CP874");
+      mbc_max = 1;
+#ifdef _MB_EXTENDED_CHARSETS_WINDOWS
+      l_wctomb = __cp_wctomb (val);
+      l_mbtowc = __cp_mbtowc (val);
+#else /* !_MB_EXTENDED_CHARSETS_WINDOWS */
+      l_wctomb = __ascii_wctomb;
+      l_mbtowc = __ascii_mbtowc;
+#endif /* _MB_EXTENDED_CHARSETS_WINDOWS */
+      break;
+#ifdef __CYGWIN__
+    /* Newlib does not provide Big5 and Cygwin's implementation
+       requires Windows support. */
     case 'B':
     case 'b':
       if (strcasecmp (charset, "BIG5"))
-      	return NULL;
+      	FAIL;
       strcpy (charset, "BIG5");
       mbc_max = 2;
-#ifdef _MB_CAPABLE
       l_wctomb = __big5_wctomb;
       l_mbtowc = __big5_mbtowc;
-#endif
       break;
 #endif /* __CYGWIN__ */
     default:
-      return NULL;
+      FAIL;
     }
-  if (category == LC_CTYPE)
+  switch (category)
     {
-      strcpy (lc_ctype_charset, charset);
-      __mb_cur_max = mbc_max;
-#ifdef _MB_CAPABLE
-      __wctomb = l_wctomb;
-      __mbtowc = l_mbtowc;
-      __set_ctype (charset);
-      /* Check for the language part of the locale specifier.  In case
-         of "ja", "ko", or "zh", assume the use of CJK fonts, unless the
-	 "@cjknarrow" modifier has been specifed.
-	 The result is stored in lc_ctype_cjk_lang and tested in wcwidth()
-	 to figure out the width to return (1 or 2) for the "CJK Ambiguous
-	 Width" category of characters. */
-      lc_ctype_cjk_lang = !cjknarrow
-			  && ((strncmp (locale, "ja", 2) == 0
-			      || strncmp (locale, "ko", 2) == 0
-			      || strncmp (locale, "zh", 2) == 0));
+    case LC_CTYPE:
+#ifndef __HAVE_LOCALE_INFO__
+      strcpy (loc->ctype_codeset, charset);
+      loc->mb_cur_max[0] = mbc_max;
 #endif
+#ifdef __CYGWIN__
+      __mb_cur_max = mbc_max;	/* Only for backward compat */
+#endif
+      loc->wctomb = l_wctomb;
+      loc->mbtowc = l_mbtowc;
+      __set_ctype (loc, charset);
+      /* Determine the width for the "CJK Ambiguous Width" category of
+         characters. This is used in wcwidth(). Assume single width for
+         single-byte charsets, and double width for multi-byte charsets
+         other than UTF-8. For UTF-8, use double width for the East Asian
+         languages ("ja", "ko", "zh"), and single width for everything else.
+         Single width can also be forced with the "@cjknarrow" modifier. */
+      loc->cjk_lang = !cjknarrow && mbc_max > 1
+		      && (charset[0] != 'U'
+			  || strncmp (locale, "ja", 2) == 0
+			  || strncmp (locale, "ko", 2) == 0
+			  || strncmp (locale, "zh", 2) == 0);
+#ifdef __HAVE_LOCALE_INFO__
+      ret = __ctype_load_locale (loc, locale, (void *) l_wctomb, charset,
+				 mbc_max);
+#endif /* __HAVE_LOCALE_INFO__ */
+      break;
+    case LC_MESSAGES:
+#ifdef __HAVE_LOCALE_INFO__
+      ret = __messages_load_locale (loc, locale, (void *) l_wctomb, charset);
+      if (!ret)
+#else
+      strcpy (loc->message_codeset, charset);
+#endif /* __HAVE_LOCALE_INFO__ */
+      break;
+#ifdef __HAVE_LOCALE_INFO__
+#ifdef __CYGWIN__
+  /* Right now only Cygwin supports a __collate_load_locale function at all. */
+    case LC_COLLATE:
+      ret = __collate_load_locale (loc, locale, (void *) l_mbtowc, charset);
+      break;
+#endif
+    case LC_MONETARY:
+      ret = __monetary_load_locale (loc, locale, (void *) l_wctomb, charset);
+      break;
+    case LC_NUMERIC:
+      ret = __numeric_load_locale (loc, locale, (void *) l_wctomb, charset);
+      break;
+    case LC_TIME:
+      ret = __time_load_locale (loc, locale, (void *) l_wctomb, charset);
+      break;
+#endif /* __HAVE_LOCALE_INFO__ */
+    default:
+      break;
     }
-  else if (category == LC_MESSAGES)
-    strcpy (lc_message_charset, charset);
-  return strcpy(current_categories[category], new_categories[category]);
+#ifdef __HAVE_LOCALE_INFO__
+  if (ret)
+    FAIL;
+#endif /* __HAVE_LOCALE_INFO__ */
+  return strcpy(loc->categories[category], new_locale);
 }
 
-static const char *
-__get_locale_env(struct _reent *p, int category)
+const char *
+__get_locale_env (struct _reent *p, int category)
 {
   const char *env;
 
@@ -749,49 +975,38 @@ __get_locale_env(struct _reent *p, int category)
 
   return env;
 }
-#endif
-
-char *
-_DEFUN_VOID(__locale_charset)
-{
-  return lc_ctype_charset;
-}
-
-char *
-_DEFUN_VOID(__locale_msgcharset)
-{
-  return lc_message_charset;
-}
+#endif /* _MB_CAPABLE */
 
 int
-_DEFUN_VOID(__locale_cjk_lang)
+_DEFUN_VOID (__locale_mb_cur_max)
 {
-  return lc_ctype_cjk_lang;
+#ifdef __HAVE_LOCALE_INFO__
+  return __get_current_ctype_locale ()->mb_cur_max[0];
+#else
+  return __get_current_locale ()->mb_cur_max[0];
+#endif
 }
 
-struct lconv *
-_DEFUN(_localeconv_r, (data), 
-      struct _reent *data)
+const char *
+__locale_ctype_ptr_l (struct __locale_t *locale)
 {
-  return (struct lconv *) &lconv;
+  return locale->ctype_ptr;
+}
+
+const char *
+__locale_ctype_ptr (void)
+{
+  return __get_current_locale ()->ctype_ptr;
 }
 
 #ifndef _REENT_ONLY
 
-#ifndef __CYGWIN__
 char *
-_DEFUN(setlocale, (category, locale),
-       int category _AND
-       _CONST char *locale)
+_DEFUN (setlocale, (category, locale),
+	int category _AND
+	_CONST char *locale)
 {
   return _setlocale_r (_REENT, category, locale);
 }
-#endif
-
-struct lconv *
-_DEFUN_VOID(localeconv)
-{
-  return _localeconv_r (_REENT);
-}
 
 #endif
